@@ -20,7 +20,6 @@ along with localreg.  If not, see <http://www.gnu.org/licenses/>.
 #!/usr/bin/env python
 
 # TBD:
-# - Multiple dependent variables
 # - Test that validation error increases
 # - Add version check on save/load
 
@@ -31,7 +30,8 @@ from . import rbf as kernels, metrics
 
 def plot_corr(axis, true, pred, log=False, *args, **kwargs):
     """
-    Visualize correlation between true and predicted values
+    Visualize correlation between true and predicted values. For multivariate
+    output, only one variable must be considered at a time.
 
     Parameters
     ----------
@@ -62,15 +62,18 @@ class RBFnet(object):
     """
     A radial basis function (RBF) machine learning network.
 
-    An arbitrary and possibly unknown function f(x) is approximated as a
-    sum of radial basis functions g,
+    A function f(x) approximates data y (possibly vector-valued) using a sum of
+    radial basis functions g:
 
-        f(x) ~ sum_i w_i g(|x-c_i|/r)
+        y_i ~ f_i(x) = sum_j w_ij g(|x-c_j|/r)
 
-    x is the input, possibly a vector. c_i and w_i are the centers and weights
-    of basis function i, while r is its radius (the same for all basis
-    functions). c_i, w_i and r are found during training, and are later used
-    for prediction.
+    x is the input, also possibly a vector. c_j is the center coordinate of
+    basis function j, while w_ij is the weight of this basis function for
+    output variable i. The center coordinate is the same for all output
+    variables y_i (for different centers per ouptut, use different RBFnet
+    objects). r is the radius, which is the same for all basis functions. This
+    is the inverse of the scaling parameter epsilon sometimes used. c_i, w_ij
+    and r are fitted during training, and are later used for predictions.
 
     The methods train() and predict() are the two most important to start with.
     """
@@ -88,20 +91,29 @@ class RBFnet(object):
         Parameters
         ----------
         input: numpy.ndarray
-            input[i,j] is prediction point i, independent variable j (x_j)
+            input[i,j] is prediction point i, independent variable j (x_j). For
+            univariate input, the last index can be omitted.
 
         Returns
         -------
-        numpy.ndarray where output[i] is the prediction of point i
+        numpy.ndarray where output[i,j] is the prediction of point i, dependent
+        variable j (y_j). The last index is omitted, if the last index were
+        omitted during training.
         """
         inp = self.normalize_input(input)
 
-        # TBD: This loop can probably be removed. For multiple output the
-        # output of this function must be a 2D array.
-        output = np.zeros(inp.shape[0])
-        for j in range(len(self.centers)):
-            distance = np.linalg.norm(inp-self.centers[j,:], axis=1)
-            output += self.coeffs[j]*self.rbf(distance/self.radius)
+        # In diff[k,i,l], 
+        # k is the index of the data point,
+        # i is the index of the basis function,
+        # l is the independent (input) variable index
+        diff = inp[:,None,:] - self.centers[None,:,:]
+
+        # distance[k,i] is then the euclidian distance between data point k and
+        # basis function i
+        distance = np.linalg.norm(diff, axis=2)
+
+        rbf_matrix = self.rbf(distance/self.radius)
+        output = rbf_matrix.dot(self.coeffs)
 
         output = self.denormalize_output(output)
 
@@ -126,12 +138,12 @@ class RBFnet(object):
         Parameters
         ----------
         input: numpy.ndarray
-            input[i,j] is data point i, independent variable j (x_j)
+            input[i,j] is data point i, independent variable j (x_j). For
+            univariate input the last index can be omitted.
 
         output: numpy.ndarray
-            output[i] is data point i, dependent variable (y). Each RBFnet object
-            only represents one dependent variable. Use multiple objects for
-            multiple dependent variables.
+            output[i,j] is data point i, dependent variable j (y_j). For
+            univariate output, the last index can be omitted.
 
         num: integer
             Number of radial basis functions (learning capacity of network).
@@ -155,11 +167,12 @@ class RBFnet(object):
             sklearn.cluster.KMeans in Scikit-Learn.
 
         keep_aspect: bool
-            Whether to scale all independent variables by the same factor
-            during normalization, i.e., to keep the aspect ratio, or to scale
-            them independently such that the standard deviation is one along
-            each axis. It may make sense to keep the aspect ratio if the input
-            variables have the same physical dimension.
+            Whether to scale all independent (input) and dependent (output)
+            variables by the same factor during normalization, i.e., to keep
+            the aspect ratio, or to scale them independently such that the
+            standard deviation is one along each axis. It may make sense to
+            keep the aspect ratio if the input variables have the same physical
+            dimension.
 
         verbose: bool
             Print progress information
@@ -208,13 +221,16 @@ class RBFnet(object):
 
         See RBFnet.train() for explanation of parameters.
         """
-        self.output_shift = np.mean(output, axis=0)
-        self.output_scale = np.std(output, axis=0)
         self.input_shift = np.mean(input, axis=0)
+        self.output_shift = np.mean(output, axis=0)
+
         if keep_aspect:
             self.input_scale = np.sqrt(np.mean(np.linalg.norm(input-self.input_shift, axis=1)**2))
+            output_ = output.reshape(len(output), 1) # Make sure it is always 2D
+            self.output_scale = np.sqrt(np.mean(np.linalg.norm(output_-self.output_shift, axis=1)**2))
         else:
             self.input_scale = np.std(input, axis=0)
+            self.output_scale = np.std(output, axis=0)
 
     def compute_centers(self, input, num, random_state=None):
         """
@@ -245,22 +261,30 @@ class RBFnet(object):
             "weights, but had {} when computing centers"\
             .format(inp.shape[1], self.centers.shape[1])
 
-        # TBD: Matrix is the same for multiple output
-        matrix = np.zeros((len(inp), len(self.centers)), dtype=float)
-        for j in range(len(self.centers)):
-            distance = np.linalg.norm(inp[:,:]-self.centers[j,:], axis=1)
-            matrix[:,j] = rbf(distance/radius)
+        # In diff[k,i,l], 
+        # k is the index of the data point,
+        # i is the index of the basis function,
+        # l is the independent (input) variable index
+        diff = inp[:,None,:] - self.centers[None,:,:]
+
+        # distance[k,i] is then the euclidian distance between data point k and
+        # basis function i
+        distance = np.linalg.norm(diff, axis=2)
+
+        # The matrix is the same for each output. The lstsq function
+        # automatically apply the least squares for each column using the same
+        # matrix.
+        rbf_matrix = rbf(distance/radius)
+
+        # TBD: This matrix is generated the same way here and in predict().
+        # Remove duplicate code.
 
-        # TBD: For multiple output, one could either consider running this for each
-        # output variable, or one could try to squeeze it into one vector and do
-        # least squares on the whole thing. That would probably lead to stacking
-        # the matrix. I wonder if it is possible to do that in a sparse way?
         if relative:
             precond = (outp+self.output_shift/self.output_scale)**(-1)
-            coeffs, _, _, _ = np.linalg.lstsq(precond[:,None]*matrix,
+            coeffs, _, _, _ = np.linalg.lstsq(precond[:,None]*rbf_matrix,
                                               precond*outp, rcond=None)
         else:
-            coeffs, _, _, _ = np.linalg.lstsq(matrix, outp, rcond=None)
+            coeffs, _, _, _ = np.linalg.lstsq(rbf_matrix, outp, rcond=None)
 
         self.coeffs = coeffs
         self.rbf = rbf
@@ -366,6 +390,8 @@ class RBFnet(object):
         """
         Evaluate each basis function separately for a sequence of input points.
 
+        NB: This function is not yet adapted to multivariate output.
+
         Parameters
         ----------
         input: numpy.ndarray
@@ -390,6 +416,8 @@ class RBFnet(object):
         """
         Plot the basis functions for a sequence of input points.
 
+        NB: This function is not yet adapted to multivariate output.
+
         Parameters
         ----------
         axis: matplotlib.axis
@@ -430,7 +458,6 @@ class RBFnet(object):
         input: numpy.array
             Array of values in the codomain
         """
-        # TBD: For multiple output this must be similar to normalize_input
         norm_output = (np.array(output)-self.output_shift)/self.output_scale
         return norm_output
 
